// Mandelbrot shader
//
// Like mandelbrot1.frag,
// plus check whether orbit is caught in cycle -> inside Mandelbrot set;
// unfortunately, this is even slower than the brute-force method (mandelbrot1)! :-(
// Hit 's' in RenderMonkey's preview window for FPS statistics ('h' = help)
//
// Requires SM4.0, i.e., GeForce 8000
// 
// Author: Gabriel Zachmann, June 2007


#version 110


uniform int MaxIterations;    // max num iterations (try: 10 / 20 / 100 / 700)
uniform vec4 InnerColor;      // two colors that the bands around (try: red)
uniform vec4 OuterColor;      // the mandelbrot set interpolate (try: blue)
uniform float BandFrequ;      // try: 0.1


vec2 cmul( vec2 c1, vec2 c2 )
{
    return vec2( c1.x * c2.x - c1.y * c2.y ,
                 c1.x * c2.y + c1.y * c2.x );
}


void main()
{
	// iterate function
	vec2 z = gl_TexCoord[0].st;

	const int num_hist = 8;
	vec2 hist[num_hist];      // history; leave it uninit'ed for now -- dangerous!
	const vec2 eps = vec2(0.0001, 0.0001);
	int j;
	for ( j = 0; j < num_hist; j ++ )
		hist[j] = z;

	// vec4 color = vec4( 0.13, 0.13, 0.13, 1.0 ); // debugging: show cycles in yellow

	bool cycle = false;
	int i;
	for ( i = 0; i < MaxIterations && !cycle; i ++ )
	{
		z = cmul(z, z) + gl_TexCoord[0].st;

		// check for cycle
		// we assume small histories, and code the test using just one branch
		int j;
		for ( j = 0; j < num_hist && !cycle; j ++ )
		{
			if ( all( lessThan( abs(z - hist[j]), eps ) ) )
			{
				// color = vec4( 0.9, 0.9, 0.13, 1.0 );    // show cycle
				cycle = true;
			}
		}

		if ( ! cycle )
		{
			// shift history
			for ( j = 1; j < num_hist; j ++ )
				hist[j-1] = hist[j];
			hist[num_hist-1] = z;
		}

		// check whether we're outside the 2-circle
		if ( dot(z,z) > 4.0 )
			break;
	}

	// color according to # iterations
	gl_FragColor = vec4( 0.13, 0.13, 0.13, 1.0 );
	// gl_FragColor = color;                // debugging: show cycles in yellow
	if ( i < MaxIterations && ! cycle )
	{
		// classical coloring with "banding"
		//float f = float(i) / float(MaxIterations);
		//gl_FragColor = mix( InnerColor, OuterColor, fract(f * BandFrequ) );
		// coloring using 'renormalized iteration count'
		// (see e.g. http://linas.org/art-gallery/escape/smooth.html )
		float f = float(i) + 1.0 - log(log(length(z)));
		gl_FragColor = mix( InnerColor, OuterColor, f * BandFrequ );
	}

	gl_FragColor *= gl_Color;
}
