#include <locale.h> // sscanf locale
#include <iostream>

#include <QFile>
#include <QTextStream>
#include "objloader.hpp"

// Very, VERY simple OBJ loader.
// Here is a short list of features a real function would provide:
// - Binary files. Reading a model should be just a few memcpy's away, not parsing a file at runtime. In short : OBJ is not very great.
// - Animations & bones (includes bones weights)
// - Multiple UVs
// - All attributes should be optional, not "forced"
// - More stable. Change a line in the OBJ file and it crashes.
// - More secure. Change another line and you can inject code.
// - Loading from memory, stream etc.

bool loadOBJ(
	const char * path,
	std::vector<QVector3D> & out_vertices,
	std::vector<QVector2D> & out_uvs,
	std::vector<QVector3D> & out_normals
)
{
	printf( "Loading OBJ file %s...\n", path );

	std::vector<unsigned int> vertexIndices, uvIndices, normalIndices;
	std::vector<QVector3D> temp_vertices;
	std::vector<QVector2D> temp_uvs;
	std::vector<QVector3D> temp_normals;

	// floating comma in sscanf depends on locale settings
	setlocale( LC_NUMERIC, "C" );

	QFile qfile( path );
	if ( !qfile.open( QIODevice::ReadOnly | QIODevice::Text ) )
	{
		std::cerr << "Impossible to open the OBJ file!" << std::endl;
		return false;
	}

	QTextStream in( &qfile );
	QString line;
	float f[3];
	unsigned int d[9];
	while ( !in.atEnd() )
	{
		// read one line of text
		line = in.readLine();
		if ( line.isEmpty() )
			continue; // EOF = End Of File. Quit the loop.

		// else parse line
		QByteArray ba = line.toUtf8(); // You cannot combine this with next line!
		const char* cline = ba.data(); // BEWARE: If QByteArray is not referenced, .data() will point at random data! (at least on Windows)
		if ( 3 == sscanf( cline, "v %f %f %f", &f[0], &f[1], &f[2] ) )
		{
			temp_vertices.push_back( QVector3D( f[0], f[1], f[2] ) );
		}
		else if ( 2 == sscanf( cline, "vt %f %f", &f[0], &f[1] ) )
		{
			// Invert V coordinate since we will only use DDS texture, which are inverted. Remove if you want to use TGA or BMP loaders.
			temp_uvs.push_back( QVector2D( f[0], -f[1] ) );
		}
		else if ( 3 == sscanf( cline, "vn %f %f %f", &f[0], &f[1], &f[2] ) )
		{
			// Invert V coordinate since we will only use DDS texture, which are inverted. Remove if you want to use TGA or BMP loaders.
			temp_normals.push_back( QVector3D( f[0], f[1], f[2] ) );
		}
		// TODO: other combinations
		else if ( 9 == sscanf( cline, "f %d/%d/%d %d/%d/%d %d/%d/%d", &d[0], &d[1], &d[2], &d[3], &d[4], &d[5], &d[6], &d[7], &d[8] ) )
		{
			vertexIndices.push_back( d[0] );
			vertexIndices.push_back( d[3] );
			vertexIndices.push_back( d[6] );
			uvIndices    .push_back( d[1] );
			uvIndices    .push_back( d[4] );
			uvIndices    .push_back( d[7] );
			normalIndices.push_back( d[2] );
			normalIndices.push_back( d[5] );
			normalIndices.push_back( d[8] );
		}
		else if ( 6 == sscanf( cline, "f %d//%d %d//%d %d//%d", &d[0], &d[1], &d[2], &d[3], &d[4], &d[5] ) )
		{
			vertexIndices.push_back( d[0] );
			vertexIndices.push_back( d[2] );
			vertexIndices.push_back( d[4] );
			normalIndices.push_back( d[1] );
			normalIndices.push_back( d[3] );
			normalIndices.push_back( d[5] );
		}
		else if ( 3 == sscanf( cline, "f %d %d %d", &d[0], &d[1], &d[2] ) )
		{
			vertexIndices.push_back( d[0] );
			vertexIndices.push_back( d[1] );
			vertexIndices.push_back( d[2] );
		}
	}
	qfile.close();

	// For each vertex of each triangle
	for ( unsigned int i = 0; i < vertexIndices.size(); ++i )
	{
		// Get the indices of its attributes
		unsigned int vertexIndex = vertexIndices[i];
		// Get the attributes thanks to the index
		QVector3D vertex = temp_vertices[ vertexIndex-1 ];
		// Put the attributes in buffers
		out_vertices.push_back(vertex);

		if ( !uvIndices.empty() )
		{
			unsigned int uvIndex = uvIndices[i];
			QVector2D uv = temp_uvs[ uvIndex-1 ];
			out_uvs.push_back(uv);
		}

		if ( !normalIndices.empty() )
		{
			unsigned int normalIndex = normalIndices[i];
			QVector3D normal = temp_normals[ normalIndex-1 ];
			out_normals.push_back(normal);
		}
	}

	return true;
}
